"""List of constants used by Eddington."""
DEFAULT_MIN_COEFF = -10.0
DEFAULT_MAX_COEFF = 10.0
DEFAULT_MEASUREMENTS = 20
DEFAULT_XMIN = -20.0
DEFAULT_XMAX = 20.0
DEFAULT_XSIGMA = 0.5
DEFAULT_YSIGMA = 0.5

DEFAULT_PRECISION = 3
